/***************************************
  C Cross Referencing & Documentation tool. Version 1.6e.

  Writes the HTML output.
  ******************/ /******************
  Written by Andrew M. Bishop

  This file Copyright 1995-2013 Andrew M. Bishop
  It may be distributed under the GNU Public License, version 2, or
  any higher version.  See section COPYING of the GNU Public license
  for conditions under which this file may be redistributed.
  ***************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "version.h"
#include "memory.h"
#include "datatype.h"
#include "cxref.h"

/*+ The file extension to use for the output files. +*/
#define HTML_FILE        ".html"
#define HTML_FILE_BACKUP ".html~"

/*+ The file extension to use for the output source files. +*/
#define HTML_SRC_FILE    ".src.html"

/*+ The name of the output tex file that contains the appendix. +*/
#define HTML_APDX        ".apdx"

/*+ The comments are to be inserted verbatim. +*/
extern int option_verbatim_comments;

/*+ The type of HTML output to produce. +*/
extern int option_html;

/*+ The name of the directory for the output. +*/
extern char* option_odir;

/*+ The base name of the file for the output. +*/
extern char* option_name;

/*+ The information about the cxref run, +*/
extern char *run_command,       /*+ the command line options. +*/
            *run_cpp_command;   /*+ the cpp command and options. +*/

/*+ The directories to go back to get to the base output directory. +*/
static char* goback=NULL;

/*+ The built-in style sheet. +*/
extern char *html_cxref_style;

/* Local functions */

static void WriteHTMLFilePart(File file);
static void WriteHTMLInclude(Include inc);
static void WriteHTMLSubInclude(Include inc,int depth);
static void WriteHTMLDefine(Define def);
static void WriteHTMLTypedef(Typedef type);
static void WriteHTMLStructUnion(StructUnion su,int depth);
static void WriteHTMLVariable(Variable var);
static void WriteHTMLFunction(Function func);

static void WriteHTMLDocument(char* name,int appendix);
static void WriteHTMLPreamble(FILE* f,char* title,int sourcefile);
static void WriteHTMLPostamble(FILE* f,int sourcefile);

void WriteHTMLSource(char *name);

static char* html(char* c,int verbatim);

/*+ The output file for the HTML. +*/
static FILE* of;

/*+ The name of the file. +*/
static char *filename;


/*++++++++++++++++++++++++++++++++++++++
  Write an html file for a complete File structure and all components.

  File file The File structure to output.
  ++++++++++++++++++++++++++++++++++++++*/

void WriteHTMLFile(File file)
{
 char* ofile;
 int i;

 filename=file->name;

 /* Write the including file. */

 goback="";
 WriteHTMLDocument(file->name,0);

 /* Open the file */

 ofile=ConcatStrings(4,option_odir,"/",file->name,HTML_FILE);

 of=fopen(ofile,"w");
 if(!of)
   {
    struct stat stat_buf;
    int i,ofl=strlen(ofile);

    for(i=strlen(option_odir)+1;i<ofl;i++)
       if(ofile[i]=='/')
         {
          ofile[i]=0;
          if(stat(ofile,&stat_buf))
             mkdir(ofile,S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
          ofile[i]='/';
         }

    of=fopen(ofile,"w");
   }

 if(!of)
   {fprintf(stderr,"cxref: Failed to open the HTML output file '%s'\n",ofile);exit(1);}

 for(goback="",i=strlen(file->name);i>0;i--)
    if(file->name[i]=='/')
       goback=ConcatStrings(2,goback,"../");

 /* Write out a header. */

 WriteHTMLPreamble(of,ConcatStrings(5,"Cross reference for ",file->name," of ",option_name,"."),0);

 /*+ The file structure is broken into its components and they are each written out. +*/

 WriteHTMLFilePart(file);

 if(file->includes)
   {
    Include inc =file->includes;
    fprintf(of,"\n<hr>\n<h2>Included Files</h2>\n\n");
    do{
       WriteHTMLInclude(inc);
      }
    while((inc=inc->next));
   }

 if(file->defines)
   {
    Define def =file->defines;
    fprintf(of,"\n<hr>\n<h2>Preprocessor definitions</h2>\n\n");
    do{
       WriteHTMLDefine(def);
      }
    while((def=def->next));
   }

 if(file->typedefs)
   {
    Typedef type=file->typedefs;
    do{
       WriteHTMLTypedef(type);
      }
    while((type=type->next));
   }

 if(file->variables)
   {
    int any_to_mention=0;
    Variable var=file->variables;

    do{
       if(var->scope&(GLOBAL|LOCAL|EXTERNAL|EXTERN_F))
          any_to_mention=1;
      }
    while((var=var->next));

    if(any_to_mention)
      {
       int first_ext=1,first_local=1;
       Variable var=file->variables;
       do{
          if(var->scope&GLOBAL)
             WriteHTMLVariable(var);
         }
       while((var=var->next));
       var=file->variables;
       do{
          if(var->scope&(EXTERNAL|EXTERN_F) && !(var->scope&GLOBAL))
            {
             if(first_ext)
               {fprintf(of,"\n<hr>\n<h2>External Variables</h2>\n\n"); first_ext=0;}
             fprintf(of,"<p>\n");
             WriteHTMLVariable(var);
            }
         }
       while((var=var->next));
       var=file->variables;
       do{
          if(var->scope&LOCAL)
            {
             if(first_local)
               {fprintf(of,"\n<hr>\n<h2>Local Variables</h2>\n\n"); first_local=0;}
             fprintf(of,"<p>\n");
             WriteHTMLVariable(var);
            }
         }
       while((var=var->next));
      }
   }

 if(file->functions)
   {
    Function func=file->functions;
    do{
       if(func->scope&(GLOBAL|EXTERNAL))
          WriteHTMLFunction(func);
      }
    while((func=func->next));
    func=file->functions;
    do{
       if(func->scope&LOCAL)
          WriteHTMLFunction(func);
      }
    while((func=func->next));
   }

 WriteHTMLPostamble(of,0);

 fclose(of);

 /* Write out the source file. */

 if(option_html&16)
    WriteHTMLSource(file->name);

 /* Clear the memory in html() */

 html(NULL,0); html(NULL,0); html(NULL,0); html(NULL,0);
}


/*++++++++++++++++++++++++++++++++++++++
  Write a File structure out.

  File file The File to output.
  ++++++++++++++++++++++++++++++++++++++*/

static void WriteHTMLFilePart(File file)
{
 int i;

 if(option_html&16)
    fprintf(of,"<h1><a name=\"file\" href=\"%s%s%s\">File %s</a></h1>\n\n",goback,file->name,HTML_SRC_FILE,html(file->name,0));
 else
    fprintf(of,"<h1><a name=\"file\">File %s</a></h1>\n\n",html(file->name,0));

 if(file->comment)
   {
    if(option_verbatim_comments)
       fprintf(of,"<pre>\n%s\n</pre>\n\n",html(file->comment,0));
    else
      {
       char *rcs1=strstr(file->comment,"$Header"),*rcs2=NULL;
       if(rcs1)
         {
          rcs2=strstr(&rcs1[1],"$");
          if(rcs2)
            {
             rcs2[0]=0;
             fprintf(of,"<p>\n<b>RCS %s</b>\n",html(&rcs1[1],0));
             rcs2[0]='$';
            }
         }
       if(rcs2)
          fprintf(of,"<p>\n<span class=\"cxref-file-comment\">%s</span>\n",html(&rcs2[2],0));
       else
          fprintf(of,"<p>\n<span class=\"cxref-file-comment\">%s</span>\n",html(file->comment,0));
      }
   }

 if(file->inc_in->n)
   {
    int i;

    fprintf(of,"<p>\n");
    fprintf(of,"<table>\n");
    for(i=0;i<file->inc_in->n;i++)
      {
       fprintf(of,"  <tr>\n");
       if(i==0)
          fprintf(of,"    <td>Included in:\n");
       else
          fprintf(of,"    <td>&nbsp;\n");
       fprintf(of,"    <td><a href=\"%s%s"HTML_FILE"#file\">%s</a>\n",goback,file->inc_in->s[i],html(file->inc_in->s[i],0));
       fprintf(of,"  </tr>\n");
      }
    fprintf(of,"</table>\n");
   }

 if(file->f_refs->n || file->v_refs->n)
   {
    fprintf(of,"<p>\n");
    fprintf(of,"<table>\n");
   }

 if(file->f_refs->n)
   {
    int others=0;

    fprintf(of,"  <tr>\n");
    fprintf(of,"    <td>References Functions:\n");

    for(i=0;i<file->f_refs->n;i++)
       if(file->f_refs->s2[i])
         {
          if(i!=others)
            {
             fprintf(of,"  <tr>\n");
             fprintf(of,"    <td>&nbsp;\n");
            }
          fprintf(of,"    <td><a href=\"%s%s"HTML_FILE"#func-%s\">%s()</a>\n",goback,file->f_refs->s2[i],file->f_refs->s1[i],html(file->f_refs->s1[i],0));
          fprintf(of,"    <td><a href=\"%s%s"HTML_FILE"#func-%s\">%s</a>\n",goback,file->f_refs->s2[i],file->f_refs->s1[i],html(file->f_refs->s2[i],0));
          fprintf(of,"  </tr>\n");
         }
       else
          others++;

    if(others)
      {
       if(i==others)
          fprintf(of,"    <td colspan=2><span class=\"cxref-reference-system\">");
       else
         {
          fprintf(of,"  </tr>\n");
          fprintf(of,"  <tr>\n");
          fprintf(of,"    <td>&nbsp;\n");
          fprintf(of,"    <td colspan=2><span class=\"cxref-reference-system\">");
         }
       for(i=0;i<file->f_refs->n;i++)
          if(!file->f_refs->s2[i])
             fprintf(of,--others?" %s(),":" %s()",html(file->f_refs->s1[i],0));
       fprintf(of,"</span>\n");
       fprintf(of,"  </tr>\n");
      }
   }

 if(file->v_refs->n)
   {
    int others=0;

    fprintf(of,"  <tr>\n");
    fprintf(of,"    <td>References Variables:\n");

    for(i=0;i<file->v_refs->n;i++)
       if(file->v_refs->s2[i])
         {
          if(i!=others)
            {
             fprintf(of,"  <tr>\n");
             fprintf(of,"    <td>&nbsp;\n");
            }
          fprintf(of,"    <td><a href=\"%s%s"HTML_FILE"#var-%s\">%s</a>\n",goback,file->v_refs->s2[i],file->v_refs->s1[i],html(file->v_refs->s1[i],0));
          fprintf(of,"    <td><a href=\"%s%s"HTML_FILE"#var-%s\">%s</a>\n",goback,file->v_refs->s2[i],file->v_refs->s1[i],html(file->v_refs->s2[i],0));
          fprintf(of,"  </tr>\n");
         }
       else
          others++;

    if(others)
      {
       if(i==others)
          fprintf(of,"    <td colspan=2><span class=\"cxref-reference-system\">");
       else
         {
          fprintf(of,"  </tr>\n");
          fprintf(of,"  <tr>\n");
          fprintf(of,"    <td>&nbsp;\n");
          fprintf(of,"    <td colspan=2><span class=\"cxref-reference-system\">");
         }
       for(i=0;i<file->v_refs->n;i++)
          if(!file->v_refs->s2[i])
             fprintf(of,--others?" %s,":" %s",html(file->v_refs->s1[i],0));
       fprintf(of,"</span>\n");
       fprintf(of,"  </tr>\n");
      }
   }

 if(file->f_refs->n || file->v_refs->n)
    fprintf(of,"</table>\n");
}


/*++++++++++++++++++++++++++++++++++++++
  Write an Include structure out.

  Include inc The Include structure to output.
  ++++++++++++++++++++++++++++++++++++++*/

static void WriteHTMLInclude(Include inc)
{
 if(inc->comment)
    fprintf(of,"<p>\n<span class=\"cxref-include-comment\">%s</span>\n<p>\n",html(inc->comment,0));

 fprintf(of,"<ul class=\"cxref-include\">\n");

 if(inc->scope==LOCAL)
    fprintf(of,"  <li class=\"cxref-include-local\"><tt><span class=\"cxref-include-local\"><a href=\"%s%s"HTML_FILE"#file\">#include \"%s\"</a></span></tt>\n",goback,inc->name,html(inc->name,0));
 else
    fprintf(of,"  <li class=\"cxref-include-system\"><tt><span class=\"cxref-include-system\">#include &lt;%s&gt;</span></tt>\n",html(inc->name,0));

 if(inc->includes)
    WriteHTMLSubInclude(inc->includes,1);

 fprintf(of,"</ul>\n");
}


/*++++++++++++++++++++++++++++++++++++++
  Write an Sub Include structure out. (An include structure that is included from another file.)

  Include inc The Include structure to output.

  int depth The depth of the include hierarchy.
  ++++++++++++++++++++++++++++++++++++++*/

static void WriteHTMLSubInclude(Include inc,int depth)
{
 int i;

 for(i=0;i<depth;i++)
    fprintf(of,"  ");
 fprintf(of,"<ul class=\"cxref-include\">\n");

 while(inc)
   {
    for(i=0;i<depth;i++)
       fprintf(of,"  ");
    if(inc->scope==LOCAL)
       fprintf(of,"  <li class=\"cxref-include-local\"><tt><span class=\"cxref-include-local\"><a href=\"%s%s"HTML_FILE"#file\">#include \"%s\"</a></span></tt>\n",goback,inc->name,html(inc->name,0));
    else
       fprintf(of,"  <li class=\"cxref-include-system\"><tt><span class=\"cxref-include-system\">#include &lt;%s&gt;</span></tt>\n",html(inc->name,0));

    if(inc->includes)
       WriteHTMLSubInclude(inc->includes,depth+1);

    inc=inc->next;
   }

 for(i=0;i<depth;i++)
    fprintf(of,"  ");
 fprintf(of,"</ul>\n");
}


/*++++++++++++++++++++++++++++++++++++++
  Write a Define structure out.

  Define def The Define structure to output.
  ++++++++++++++++++++++++++++++++++++++*/

static void WriteHTMLDefine(Define def)
{
 int i;
 int pargs=0;

 if(def->comment)
    fprintf(of,"<p>\n<span class=\"cxref-define-comment\">%s</span>\n",html(def->comment,0));

 if(option_html&16)
    fprintf(of,"<p>\n<tt><span class=\"cxref-define\"><a href=\"%s%s%s#line%d\">#define %s</a>",goback,filename,HTML_SRC_FILE,def->lineno,html(def->name,0));
 else
    fprintf(of,"<p>\n<tt><span class=\"cxref-define\">#define %s",html(def->name,0));

 if(def->value)
    fprintf(of," %s",html(def->value,0));

 if(def->args->n)
   {
    fprintf(of,"( ");
    for(i=0;i<def->args->n;i++)
       fprintf(of,i?", %s":"%s",html(def->args->s1[i],0));
    fprintf(of," )");
   }
 fprintf(of,"</span></tt>\n");

 for(i=0;i<def->args->n;i++)
    if(def->args->s2[i])
       pargs=1;

 if(pargs)
   {
    fprintf(of,"<br>\n");
    fprintf(of,"<dl>\n");
    for(i=0;i<def->args->n;i++)
      {
       fprintf(of,"  <dt><tt><span class=\"cxref-define\">%s</span></tt>\n",html(def->args->s1[i],0));
       fprintf(of,"  <dd><span class=\"cxref-define-comment\">%s</span>\n",def->args->s2[i]?html(def->args->s2[i],0):"");
      }
    fprintf(of,"</dl>\n");
   }
}


/*++++++++++++++++++++++++++++++++++++++
  Write a Typedef structure out.

  Typedef type The Typedef structure to output.
  ++++++++++++++++++++++++++++++++++++++*/

static void WriteHTMLTypedef(Typedef type)
{
 fprintf(of,"\n<hr>\n<h2>");

 if(!strncmp("enum",type->name,4))
    fprintf(of,"<a name=\"type-enum-%s\">",&type->name[5]);
 else if(!strncmp("union",type->name,5))
    fprintf(of,"<a name=\"type-union-%s\">",&type->name[6]);
 else if(!strncmp("struct",type->name,6))
    fprintf(of,"<a name=\"type-struct-%s\">",&type->name[7]);
 else
    fprintf(of,"<a name=\"type-%s\">",type->name);

 if(type->type)
    fprintf(of,"Typedef %s",html(type->name,0));
 else
    fprintf(of,"Type %s",html(type->name,0));

 fprintf(of,"</a></h2>\n\n");

 if(type->comment)
    fprintf(of,"<p>\n<span class=\"cxref-type-comment\">%s</span>\n",html(type->comment,0));

 if(type->type)
   {
    if(option_html&16)
       fprintf(of,"<p>\n<tt><span class=\"cxref-type\"><a href=\"%s%s%s#line%d\">typedef %s</a></span></tt>\n",goback,filename,HTML_SRC_FILE,type->lineno,html(type->type,0));
    else
       fprintf(of,"<p>\n<tt><span class=\"cxref-type\">typedef %s</span></tt>\n",html(type->type,0));
   }
 else if(type->sutype)
   {
    if(option_html&16)
       fprintf(of,"<br>\n<tt><span class=\"cxref-type\"><a href=\"%s%s%s#line%d\">%s</a></span></tt>\n",goback,filename,HTML_SRC_FILE,type->lineno,html(type->sutype->name,0));
    else
       fprintf(of,"<br>\n<tt><span class=\"cxref-type\">%s</span></tt>\n",html(type->sutype->name,0));
   }

 if(type->sutype)
   {
    fprintf(of,"<br>\n");
    fprintf(of,"<table>\n");
    WriteHTMLStructUnion(type->sutype,0);
    fprintf(of,"</table>\n");
   }
 else
    if(type->typexref)
      {
       fprintf(of,"<br>\n");
       fprintf(of,"<table>\n");
       fprintf(of,"  <tr>\n");
       fprintf(of,"    <td>See:\n");
       if(type->typexref->type)
          fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"#type-%s\">Typedef %s</a></span>\n",type->typexref->name,html(type->typexref->name,0));
       else if(!strncmp("enum",type->typexref->name,4))
          fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"#type-enum-%s\">Type %s</a></span>\n",&type->typexref->name[5],html(type->typexref->name,0));
       else if(!strncmp("union",type->typexref->name,5))
          fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"#type-union-%s\">Type %s</a></span>\n",&type->typexref->name[6],html(type->typexref->name,0));
       else if(!strncmp("struct",type->typexref->name,6))
          fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"#type-struct-%s\">Type %s</a></span>\n",&type->typexref->name[7],html(type->typexref->name,0));
       fprintf(of,"  </tr>\n");
       fprintf(of,"</table>\n");
      }
}


/*++++++++++++++++++++++++++++++++++++++
  Write a structure / union structure out.

  StructUnion su The structure / union to write.

  int depth The current depth within the structure.
  ++++++++++++++++++++++++++++++++++++++*/

static void WriteHTMLStructUnion(StructUnion su, int depth)
{
 int i;
 char* splitsu=NULL;

 splitsu=strstr(su->name,"{...}");
 if(splitsu) splitsu[-1]=0;

 fprintf(of,"  <tr>\n");
 fprintf(of,"    <td>");
 for(i=0;i<depth;i++)
    fprintf(of,"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\n");
 if(!depth || su->comps)
    fprintf(of,"<tt><span class=\"cxref-type\">%s</span></tt>\n",html(su->name,0));
 else
    fprintf(of,"<tt><span class=\"cxref-type\">%s;</span></tt>\n",html(su->name,0));
 fprintf(of,"    <td>");
 if(depth && su->comment && !su->comps)
    fprintf(of,"<span class=\"cxref-type-comment\">%s</span>\n",html(su->comment,0));
 else
    fprintf(of,"&nbsp;\n");
 fprintf(of,"  </tr>\n");

 if(!depth || su->comps)
   {
    fprintf(of,"  <tr>\n");
    fprintf(of,"    <td>");
    for(i=0;i<depth;i++)
       fprintf(of,"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
    fprintf(of,"&nbsp;&nbsp;&nbsp;<tt><span class=\"cxref-type\">{</span></tt>\n");
    fprintf(of,"    <td>&nbsp;\n");
    fprintf(of,"  </tr>\n");

    for(i=0;i<su->n_comp;i++)
       WriteHTMLStructUnion(su->comps[i],depth+1);

    fprintf(of,"  <tr>\n");
    fprintf(of,"    <td>");
    for(i=0;i<depth;i++)
       fprintf(of,"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
    fprintf(of,"&nbsp;&nbsp;&nbsp;<tt><span class=\"cxref-type\">}</span></tt>\n");
    fprintf(of,"    <td>&nbsp;\n");
    fprintf(of,"  </tr>\n");

    if(splitsu)
      {
       fprintf(of,"  <tr>\n");
       fprintf(of,"    <td>");
       for(i=0;i<depth;i++)
          fprintf(of,"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
       fprintf(of,"<tt><span class=\"cxref-type\">%s;</span></tt>\n",splitsu[5]?html(&splitsu[6],0):"");
       if(depth && su->comment)
          fprintf(of,"    <td><span class=\"cxref-type-comment\">%s</span>\n",html(su->comment,0));
       else
          fprintf(of,"    <td>&nbsp;\n");
       fprintf(of,"  </tr>\n");
      }
   }

 if(splitsu) splitsu[-1]=' ';
}


/*++++++++++++++++++++++++++++++++++++++
  Write a Variable structure out.

  Variable var The Variable structure to output.
  ++++++++++++++++++++++++++++++++++++++*/

static void WriteHTMLVariable(Variable var)
{
 int i;

 if(var->scope&GLOBAL)
    fprintf(of,"\n<hr>\n<h2><a name=\"var-%s\">Global Variable %s</a></h2>\n\n",var->name,html(var->name,0));
 else
    fprintf(of,"<b><a name=\"var-%s\">%s</a></b><br>\n\n",var->name,html(var->name,0));

 if(var->comment)
    fprintf(of,"<p>\n<span class=\"cxref-variable-comment\">%s</span>\n",html(var->comment,0));

 if(option_html&16 && var->scope&(GLOBAL|LOCAL))
   {
    if(var->incfrom)
       fprintf(of,"<p>\n<tt><span class=\"cxref-variable\"><a href=\"%s%s%s#line%d\">",goback,var->incfrom,HTML_SRC_FILE,var->lineno);
    else
       fprintf(of,"<p>\n<tt><span class=\"cxref-variable\"><a href=\"%s%s%s#line%d\">",goback,filename,HTML_SRC_FILE,var->lineno);
   }
 else
    fprintf(of,"<p>\n<tt><span class=\"cxref-variable\">");

 if(var->scope&LOCAL)
    fprintf(of,"static ");
 else
    if(!(var->scope&GLOBAL) && var->scope&(EXTERNAL|EXTERN_F))
       fprintf(of,"extern ");

 fprintf(of,"%s",html(var->type,0));

 if(option_html&16 && var->scope&(GLOBAL|LOCAL))
    fprintf(of,"</a></span></tt>\n");
 else
    fprintf(of,"</span></tt>\n");

 if(var->scope&(GLOBAL|LOCAL))
   {
    if(var->incfrom || var->visible->n || var->used->n)
      {
       fprintf(of,"<br>\n");
       fprintf(of,"<table>\n");
      }

    if(var->incfrom)
      {
       fprintf(of,"  <tr>\n");
       fprintf(of,"    <td>Included from\n");
       fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#var-%s\">%s</a></span>\n",goback,var->incfrom,var->name,html(var->incfrom,0));
       fprintf(of,"  </tr>\n");
      }

    if(var->visible->n)
      {
       for(i=0;i<var->visible->n;i++)
         {
          if(i==0)
            {
             fprintf(of,"  <tr>\n");
             fprintf(of,"    <td>Visible in:\n");
            }
          else
            {
             fprintf(of,"  <tr>\n");
             fprintf(of,"    <td>&nbsp;\n");
            }
          if(var->visible->s1[i][0]=='$' && !var->visible->s1[i][1])
            {
             fprintf(of,"    <td>&nbsp;\n");
             fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#file\">%s</a></span>\n",goback,var->visible->s2[i],html(var->visible->s2[i],0));
            }
          else
            {
             fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#func-%s\">%s()</a></span>\n",goback,var->visible->s2[i],var->visible->s1[i],html(var->visible->s1[i],0));
             fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#func-%s\">%s</a></span>\n",goback,var->visible->s2[i],var->visible->s1[i],html(var->visible->s2[i],0));
            }
          fprintf(of,"  </tr>\n");
         }
      }

    if(var->used->n)
      {
       for(i=0;i<var->used->n;i++)
         {
          if(i==0)
            {
             fprintf(of,"  <tr>\n");
             fprintf(of,"    <td>Used in:\n");
            }
          else
            {
             fprintf(of,"  <tr>\n");
             fprintf(of,"    <td>&nbsp;\n");
            }
          if(var->used->s1[i][0]=='$' && !var->used->s1[i][1])
             fprintf(of,"    <td>&nbsp;    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#file\">%s</a></span>\n",goback,var->used->s2[i],html(var->used->s2[i],0));
          else
            {
             if(var->scope&LOCAL)
               {
                fprintf(of,"    <td><a href=\"#func-%s\">%s()</a>\n",var->used->s1[i],html(var->used->s1[i],0));
                fprintf(of,"    <td>&nbsp;\n");
               }
             else
               {
                fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#func-%s\">%s()</a></span>\n",goback,var->used->s2[i],var->used->s1[i],html(var->used->s1[i],0));
                fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#func-%s\">%s</a></span>\n",goback,var->used->s2[i],var->used->s1[i],html(var->used->s2[i],0));
               }
            }
          fprintf(of,"  </tr>\n");
         }
      }

    if(var->incfrom || var->visible->n || var->used->n)
       fprintf(of,"</table>\n");
   }
 else
    if(var->scope&(EXTERNAL|EXTERN_F) && var->defined)
      {
       fprintf(of,"<table>\n");
       fprintf(of,"  <tr>\n");
       fprintf(of,"    <td>Defined in:\n");
       fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#var-%s\">%s</a></span>\n",goback,var->defined,html(var->name,0),var->defined);
       fprintf(of,"  </tr>\n");
       fprintf(of,"</table>\n");
      }
}


/*++++++++++++++++++++++++++++++++++++++
  Write a Function structure out.

  Function func The Function structure to output.
  ++++++++++++++++++++++++++++++++++++++*/

static void WriteHTMLFunction(Function func)
{
 int i,pret,pargs;
 char* comment2=NULL,*type;

 if(func->scope&(GLOBAL|EXTERNAL))
    fprintf(of,"\n<hr>\n<h2><a name=\"func-%s\">Global Function %s()</a></h2>\n\n",func->name,html(func->name,0));
 else
    fprintf(of,"\n<hr>\n<h2><a name=\"func-%s\">Local Function %s()</a></h2>\n\n",func->name,html(func->name,0));

 if(func->comment)
   {
    if(option_verbatim_comments)
       fprintf(of,"<pre>\n%s\n</pre>\n\n",html(func->comment,0));
    else
      {
       comment2=strstr(func->comment,"\n\n");
       if(comment2)
          comment2[0]=0;
       fprintf(of,"<p>\n<span class=\"cxref-function-comment\">%s</span>\n",html(func->comment,0));
      }
   }

 if(option_html&16)
   {
    if(func->incfrom)
       fprintf(of,"<p>\n<tt><span class=\"cxref-function\"><a href=\"%s%s%s#line%d\">",goback,func->incfrom,HTML_SRC_FILE,func->lineno);
    else
       fprintf(of,"<p>\n<tt><span class=\"cxref-function\"><a href=\"%s%s%s#line%d\">",goback,filename,HTML_SRC_FILE,func->lineno);
   }
 else
    fprintf(of,"<p>\n<tt><span class=\"cxref-function\">");

 if(func->scope&LOCAL)
    fprintf(of,"static ");
 if(func->scope&INLINED)
   fprintf(of,"inline ");

 if((type=strstr(func->type,"()")))
    type[0]=0;
 fprintf(of,"%s ( ",html(func->type,0));

 for(i=0;i<func->args->n;i++)
    fprintf(of,i?", %s":"%s",html(func->args->s1[i],0));

 if(type)
   {fprintf(of," %s",html(&type[1],0));type[0]='(';}
 else
    fprintf(of," )");

 if(option_html&16)
    fprintf(of,"</a></span></tt>\n");
 else
    fprintf(of,"</span></tt>\n");

 pret =strncmp("void ",func->type,5) && func->cret;
 for(pargs=0,i=0;i<func->args->n;i++)
    pargs = pargs || ( strcmp("void",func->args->s1[i]) && func->args->s2[i] );

 if(pret || pargs)
   {
    fprintf(of,"<br>\n");
    fprintf(of,"<dl>\n");
    if(pret)
      {
       fprintf(of,"  <dt><tt><span class=\"cxref-function\">%s</span></tt>\n",html(func->type,0));
       fprintf(of,"  <dd><span class=\"cxref-function-comment\">%s</span>\n",func->cret?html(func->cret,0):"&nbsp;");
      }
    if(pargs)
       for(i=0;i<func->args->n;i++)
         {
          fprintf(of,"  <dt><tt><span class=\"cxref-function\">%s</span></tt>\n",html(func->args->s1[i],0));
          fprintf(of,"  <dd><span class=\"cxref-function-comment\">%s</span>\n",func->args->s2[i]?html(func->args->s2[i],0):"&nbsp;");
         }
    fprintf(of,"</dl>\n");
   }

 if(comment2)
   {
    fprintf(of,"<p>\n<span class=\"cxref-function-comment\">%s</span>\n",html(&comment2[2],0));
    comment2[0]='\n';
   }

 if(func->protofile || func->incfrom || func->calls->n || func->called->n || func->used->n || func->f_refs->n || func->v_refs->n)
   {
    fprintf(of,"<p>\n");
    fprintf(of,"<table>\n");
   }

 if(func->protofile)
   {
    fprintf(of,"  <tr>\n");
    fprintf(of,"    <td>Prototyped in:\n");
    fprintf(of,"    <td colspan=2><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#file\">%s</a></span>\n",goback,func->protofile,html(func->protofile,0));
    fprintf(of,"  </tr>\n");
   }

 if(func->incfrom)
   {
    fprintf(of,"  <tr>\n");
    fprintf(of,"    <td>Included from:\n");
    fprintf(of,"    <td colspan=2><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#func-%s\">%s</a></span>\n",goback,func->incfrom,func->name,html(func->incfrom,0));
    fprintf(of,"  </tr>\n");
   }

 if(func->calls->n)
   {
    int others=0;

    fprintf(of,"  <tr>\n");
    fprintf(of,"    <td>Calls:\n");

    for(i=0;i<func->calls->n;i++)
       if(func->calls->s2[i])
         {
          if(i!=others)
            {
             fprintf(of,"  <tr>\n");
             fprintf(of,"    <td>&nbsp;\n");
            }
          fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#func-%s\">%s()</a></span>\n",goback,func->calls->s2[i],func->calls->s1[i],html(func->calls->s1[i],0));
          fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#func-%s\">%s</a></span>\n",goback,func->calls->s2[i],func->calls->s1[i],html(func->calls->s2[i],0));
          fprintf(of,"  </tr>\n");
         }
       else
          others++;

    if(others)
      {
       if(i==others)
          fprintf(of,"    <td colspan=2><span class=\"cxref-reference-system\">");
       else
         {
          fprintf(of,"  <tr>\n");
          fprintf(of,"    <td>&nbsp;\n");
          fprintf(of,"    <td colspan=2><span class=\"cxref-reference-system\">");
         }
       for(i=0;i<func->calls->n;i++)
          if(!func->calls->s2[i])
             fprintf(of,--others?"%s(), ":"%s()",html(func->calls->s1[i],0));
       fprintf(of,"</span>\n");
       fprintf(of,"  </tr>\n");
      }
   }

 if(func->called->n)
   {
    fprintf(of,"  <tr>\n");
    fprintf(of,"    <td>Called by:\n");

    for(i=0;i<func->called->n;i++)
      {
       if(i!=0)
         {
          fprintf(of,"  <tr>\n");
          fprintf(of,"    <td>&nbsp;\n");
         }
       fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#func-%s\">%s()</a></span>\n",goback,func->called->s2[i],func->called->s1[i],html(func->called->s1[i],0));
       fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#func-%s\">%s</a></span>\n",goback,func->called->s2[i],func->called->s1[i],html(func->called->s2[i],0));
       fprintf(of,"  </tr>\n");
      }
   }

 if(func->used->n)
   {
    fprintf(of,"  <tr>\n");
    fprintf(of,"    <td>Used in:\n");

    for(i=0;i<func->used->n;i++)
      {
       if(i!=0)
         {
          fprintf(of,"  <tr>\n");
          fprintf(of,"    <td>&nbsp;\n");
         }
       if(func->used->s1[i][0]=='$' && !func->used->s1[i][1])
         {
          fprintf(of,"    <td>&nbsp;\n");
          fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#file\">%s</a></span>\n",goback,func->used->s2[i],html(func->used->s2[i],0));
         }
       else
         {
          fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#func-%s\">%s()</a></span>\n",goback,func->used->s2[i],func->used->s1[i],html(func->used->s1[i],0));
          fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#func-%s\">%s</a></span>\n",goback,func->used->s2[i],func->used->s1[i],html(func->used->s2[i],0));
         }
       fprintf(of,"  </tr>\n");
      }
   }

 if(func->f_refs->n)
   {
    int others=0;

    fprintf(of,"  <tr>\n");
    fprintf(of,"    <td>References Functions:\n");

    for(i=0;i<func->f_refs->n;i++)
       if(func->f_refs->s2[i])
         {
          if(i!=others)
            {
             fprintf(of,"  <tr>\n");
             fprintf(of,"    <td>&nbsp;\n");
            }
          fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#func-%s\">%s()</a></span>\n",goback,func->f_refs->s2[i],func->f_refs->s1[i],html(func->f_refs->s1[i],0));
          fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#func-%s\">%s</a></span>\n",goback,func->f_refs->s2[i],func->f_refs->s1[i],html(func->f_refs->s2[i],0));
          fprintf(of,"  </tr>\n");
         }
       else
          others++;

    if(others)
      {
       if(i==others)
          fprintf(of,"    <td colspan=2><span class=\"cxref-reference-system\">");
       else
         {
          fprintf(of,"  <tr>\n");
          fprintf(of,"    <td>&nbsp;\n");
          fprintf(of,"    <td colspan=2><span class=\"cxref-reference-system\">");
         }
       for(i=0;i<func->f_refs->n;i++)
          if(!func->f_refs->s2[i])
             fprintf(of,--others?"%s(), ":"%s()",html(func->f_refs->s1[i],0));
       fprintf(of,"</span>\n");
       fprintf(of,"  </tr>\n");
      }
   }

 if(func->v_refs->n)
   {
    int others=0;

    fprintf(of,"  <tr>\n");
    fprintf(of,"    <td>References Variables:\n");

    for(i=0;i<func->v_refs->n;i++)
       if(func->v_refs->s2[i])
         {
          if(i!=others)
            {
             fprintf(of,"  <tr>\n");
             fprintf(of,"    <td>&nbsp;\n");
            }
          fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#var-%s\">%s</a></span>\n",goback,func->v_refs->s2[i],func->v_refs->s1[i],html(func->v_refs->s1[i],0));
          fprintf(of,"    <td><span class=\"cxref-reference-local\"><a href=\"%s%s"HTML_FILE"#var-%s\">%s</a></span>\n",goback,func->v_refs->s2[i],func->v_refs->s1[i],html(func->v_refs->s2[i],0));
          fprintf(of,"  </tr>\n");
         }
       else
          others++;

    if(others)
      {
       if(i==others)
          fprintf(of,"    <td colspan=2><span class=\"cxref-reference-system\">");
       else
         {
          fprintf(of,"  <tr>\n");
          fprintf(of,"    <td>&nbsp;\n");
          fprintf(of,"    <td colspan=2><span class=\"cxref-reference-system\">");
         }
       for(i=0;i<func->v_refs->n;i++)
          if(!func->v_refs->s2[i])
             fprintf(of,--others?"%s, ":"%s",html(func->v_refs->s1[i],0));
       fprintf(of,"</span>\n");
       fprintf(of,"  </tr>\n");
      }
   }

 if(func->protofile || func->incfrom || func->calls->n || func->called->n || func->used->n || func->f_refs->n || func->v_refs->n)
    fprintf(of,"</table>\n");
}


/*++++++++++++++++++++++++++++++++++++++
  Write out a file that will include the current information.

  char* name The name of the file.

  int appendix set to non-zero if the appendix file is to be added, else a normal source file.  
  ++++++++++++++++++++++++++++++++++++++*/

static void WriteHTMLDocument(char* name,int appendix)
{
 FILE *in,*out;
 char line[256];
 int seen=0;
 char *inc_file,*ofile,*ifile;
 char *fname;
 struct stat stat_buf;

 if(appendix)
    inc_file=ConcatStrings(4,"<a href=\"",name,HTML_FILE,"\">Appendix</a><br>\n");
 else
    inc_file=ConcatStrings(6,"<a href=\"",name,HTML_FILE,"#file\">",name,"</a><br>\n");
 ifile=ConcatStrings(4,option_odir,"/",option_name,HTML_FILE);
 ofile=ConcatStrings(4,option_odir,"/",option_name,HTML_FILE_BACKUP);

 in=fopen(ifile,"r");
 if(!in)
   {
    in =fopen(ifile,"w");
    if(!in)
      {fprintf(stderr,"cxref: Failed to open the main HTML output file '%s'\n",ifile);exit(1);}

    WriteHTMLPreamble(in,ConcatStrings(3,"Cross Reference Of ",option_name,"."),1);
    WriteHTMLPostamble(in,1);
    fclose(in);

    in =fopen(ifile,"r");
   }

 out=fopen(ofile,"w");

 if(!out)
   {fprintf(stderr,"cxref: Failed to open the main HTML output file '%s'\n",ofile);exit(1);}

 while(fgets(line,256,in))
   {
    if(!strcmp(inc_file,line) ||
       (!strncmp("<!--",line,4) && !strncmp(inc_file,line+4,strlen(inc_file))) ||
       (!strncmp("<!-- ",line,5) && !strncmp(inc_file,line+5,strlen(inc_file))))
      {seen=1;break;}
    if(line[0]=='<' && !strcmp("<!-- End-Of-Source-Files -->\n",line))
      {
       if(appendix)
         {
          fputs(line,out);
          fputs("\n",out);
          fputs("<!-- Appendix -->\n",out);
          fputs("\n",out);
          fputs("<hr>\n",out);
          fputs("<h1>Appendix</h1>\n",out);
          fputs("\n",out);
          fputs(inc_file,out);
         }
       else
         {
          fputs(inc_file,out);
          fputs("\n",out);
          fputs(line,out);
         }
      }
    else
       fputs(line,out);
   }

 fclose(in);
 fclose(out);

 if(!seen)
   {
    unlink(ifile);
    rename(ofile,ifile);
   }
 else
    unlink(ofile);

 /* Style file */

 fname=ConcatStrings(2,option_odir,"/cxref.css");
 if(stat(fname,&stat_buf))
   {
    FILE* file=fopen(fname,"w");
    if(!file)
      {fprintf(stderr,"cxref: Cannot write the HTML style file '%s'\n",fname);exit(1);}
    fputs(html_cxref_style,file);
    fclose(file);
   }
}


/*++++++++++++++++++++++++++++++++++++++
  Write out a standard pre-amble.

  FILE* f The file to write the pre amble to.

  char* title The title of the file.

  int sourcefile True if the Source-Files line is to be included.
  ++++++++++++++++++++++++++++++++++++++*/

static void WriteHTMLPreamble(FILE* f,char* title,int sourcefile)
{
 fputs("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\n",f);
 fputs("\n",f);
 fputs("<!-- This HTML file generated by cxref (version " CXREF_VERSION "). -->\n",f);
 fputs("<!-- cxref program " CXREF_COPYRIGHT ". -->\n",f);
 fputs("\n",f);
 if(!sourcefile)
   {
    fputs("<!--\n",f);
    if(filename)
       fprintf(f,"Cxref: %s %s\n",run_command,filename);
    else
       fprintf(f,"Cxref: %s\n",run_command);
    fprintf(f,"CPP  : %s\n",run_cpp_command);
    fputs("-->\n",f);
    fputs("\n",f);
   }
 fputs("<HTML>\n",f);
 fputs("\n",f);
 fputs("<HEAD>\n",f);
 fprintf(f,"<TITLE>%s</TITLE>\n",title);
 fprintf(f,"<LINK rel=\"stylesheet\" href=\"%scxref.css\" type=\"text/css\">\n",goback);
 fputs("</HEAD>\n",f);
 fputs("\n",f);
 fputs("<BODY>\n",f);
 fputs("\n",f);
 if(sourcefile)
   {
    fputs("<h1>Source Files</h1>\n",f);
    fputs("\n",f);
    fputs("<!-- Begin-Of-Source-Files -->\n",f);
   }
}


/*++++++++++++++++++++++++++++++++++++++
  Write out a standard post-amble. This includes the end of document marker.

  FILE* f The file to write the post amble to.

  int sourcefile True if the Source-Files line is to be included.
  ++++++++++++++++++++++++++++++++++++++*/

static void WriteHTMLPostamble(FILE* f,int sourcefile)
{
 if(sourcefile)
   {
    fputs("\n",f);
    fputs("<!-- End-Of-Source-Files -->\n",f);
   }
 fputs("\n",f);
 fputs("</BODY>\n",f);
 fputs("</HTML>\n",f);
}


/*++++++++++++++++++++++++++++++++++++++
  Write out the appendix information.

  StringList files The list of files to write.

  StringList2 funcs The list of functions to write.

  StringList2 vars The list of variables to write.

  StringList2 types The list of types to write.
  ++++++++++++++++++++++++++++++++++++++*/

void WriteHTMLAppendix(StringList files,StringList2 funcs,StringList2 vars,StringList2 types)
{
 char* ofile;
 int i;

 filename=NULL;

 /* Write the bits to the including file. */

 goback="";
 WriteHTMLDocument(ConcatStrings(2,option_name,HTML_APDX),1);

 /* Open the file */

 ofile=ConcatStrings(5,option_odir,"/",option_name,HTML_APDX,HTML_FILE);

 of=fopen(ofile,"w");

 if(!of)
   {fprintf(stderr,"cxref: Failed to open the HTML appendix file '%s'\n",ofile);exit(1);}

 /* Write the file structure out */

 WriteHTMLPreamble(of,ConcatStrings(3,"Cross reference index of ",option_name,"."),0);

 fprintf(of,"<h1>Cross References</h1>\n\n");

 if(files->n || funcs->n || vars->n || types->n) 
   {
    fprintf(of,"<p>\n");
    fprintf(of,"<ul>\n");
    if(files->n) 
       fprintf(of,"  <li><a href=\"#files\">Files</a></li>\n");
    if(funcs->n) 
       fprintf(of,"  <li><a href=\"#functions\">Global Functions</a></li>\n");
    if(vars->n) 
       fprintf(of,"  <li><a href=\"#variables\">Global Variables</a></li>\n");
    if(types->n) 
       fprintf(of,"  <li><a href=\"#types\">Defined Types</a></li>\n");
    fprintf(of,"</ul>\n");
   }

 /* Write out the appendix of files. */

 if(files->n)
   {
    fprintf(of,"\n<hr>\n<h2><a name=\"files\">Files</a></h2>\n\n");
    fprintf(of,"<p>\n");
    fprintf(of,"<ul>\n");
    for(i=0;i<files->n;i++)
       fprintf(of,"  <li><a href=\"%s"HTML_FILE"#file\">%s</a></li>\n",files->s[i],html(files->s[i],0));
    fprintf(of,"</ul>\n");
   }

 /* Write out the appendix of functions. */

 if(funcs->n)
   {
    fprintf(of,"\n<hr>\n<h2><a name=\"functions\">Global Functions</a></h2>\n\n");
    fprintf(of,"<p>\n");
    fprintf(of,"<ul>\n");
    for(i=0;i<funcs->n;i++)
       fprintf(of,"  <li><a href=\"%s"HTML_FILE"#func-%s\">%s()  :  %s</a></li>\n",funcs->s2[i],funcs->s1[i],html(funcs->s1[i],0),html(funcs->s2[i],0));
    fprintf(of,"</ul>\n");
   }

 /* Write out the appendix of variables. */

 if(vars->n)
   {
    fprintf(of,"\n<hr>\n<h2><a name=\"variables\">Global Variables</a></h2>\n\n");
    fprintf(of,"<p>\n");
    fprintf(of,"<ul>\n");
    for(i=0;i<vars->n;i++)
       fprintf(of,"  <li><a href=\"%s"HTML_FILE"#var-%s\">%s  :  %s</a></li>\n",vars->s2[i],vars->s1[i],html(vars->s1[i],0),html(vars->s2[i],0));
    fprintf(of,"</ul>\n");
   }

 /* Write out the appendix of types. */

 if(types->n)
   {
    fprintf(of,"\n<hr>\n<h2><a name=\"types\">Defined Types</a></h2>\n\n");
    fprintf(of,"<p>\n");
    fprintf(of,"<ul>\n");
    for(i=0;i<types->n;i++)
       if(!strncmp("enum",types->s1[i],4))
          fprintf(of,"  <li><a href=\"%s"HTML_FILE"#type-enum-%s\">%s  :  %s</a></li>\n",types->s2[i],&types->s1[i][5],html(types->s1[i],0),html(types->s2[i],0));
       else if(!strncmp("union",types->s1[i],5))
          fprintf(of,"  <li><a href=\"%s"HTML_FILE"#type-union-%s\">%s  :  %s</a></li>\n",types->s2[i],&types->s1[i][6],html(types->s1[i],0),html(types->s2[i],0));
       else if(!strncmp("struct",types->s1[i],6))
          fprintf(of,"  <li><a href=\"%s"HTML_FILE"#type-struct-%s\">%s  :  %s</a></li>\n",types->s2[i],&types->s1[i][7],html(types->s1[i],0),html(types->s2[i],0));
       else
          fprintf(of,"  <li><a href=\"%s"HTML_FILE"#type-%s\">%s  :  %s</a></li>\n",types->s2[i],types->s1[i],html(types->s1[i],0),html(types->s2[i],0));
    fprintf(of,"</ul>\n");
   }

 WriteHTMLPostamble(of,0);

 fclose(of);

 /* Clear the memory in html(,0) */

 html(NULL,0); html(NULL,0); html(NULL,0); html(NULL,0);
}


/*++++++++++++++++++++++++++++++++++++++
  Delete the HTML file and main file reference that belong to the named file.

  char *name The name of the file to delete.
  ++++++++++++++++++++++++++++++++++++++*/

void WriteHTMLFileDelete(char *name)
{
 FILE *in,*out;
 char line[256];
 int seen=0;
 char *inc_file,*ofile,*ifile;

 ofile=ConcatStrings(4,option_odir,"/",name,HTML_FILE);
 unlink(ofile);

 inc_file=ConcatStrings(6,"<a href=\"",name,HTML_FILE,"#file\">",name,"</a><br>\n");
 ifile=ConcatStrings(4,option_odir,"/",option_name,HTML_FILE);
 ofile=ConcatStrings(4,option_odir,"/",option_name,HTML_FILE_BACKUP);

 in =fopen(ifile,"r");
 out=fopen(ofile,"w");

 if(in && !out)
   {fprintf(stderr,"cxref: Failed to open the main HTML output file '%s'\n",ofile);fclose(in);}
 else if(in)
   {
    while(fgets(line,256,in))
      {
       if(!strcmp(inc_file,line) ||
          (!strncmp("<!--",line,4) && !strncmp(inc_file,line+4,strlen(inc_file)-1)) ||
          (!strncmp("<!-- ",line,5) && !strncmp(inc_file,line+5,strlen(inc_file)-1)))
          seen=1;
       else
          fputs(line,out);
      }

    fclose(in);
    fclose(out);

    if(seen)
      {
       unlink(ifile);
       rename(ofile,ifile);
      }
    else
       unlink(ofile);
   }
 else if(out)
   {
    fclose(out);
    unlink(ofile);
   }
}


/*++++++++++++++++++++++++++++++++++++++
  Write out the source file.

  char *name The name of the source file.
  ++++++++++++++++++++++++++++++++++++++*/

void WriteHTMLSource(char *name)
{
 FILE *in,*out;
 char line[256];
 char *ofile,*ifile;
 int lineno=0;
 char pad[5];

 ifile=name;
 ofile=ConcatStrings(4,option_odir,"/",name,HTML_SRC_FILE);

 in =fopen(ifile,"r");
 if(!in)
   {fprintf(stderr,"cxref: Failed to open the source file '%s'\n",ifile);exit(1);}

 out=fopen(ofile,"w");
 if(!out)
   {fprintf(stderr,"cxref: Failed to open the HTML output source file '%s'\n",ofile);exit(1);}

 WriteHTMLPreamble(out,ConcatStrings(2,"Source File ",name),0);
 fputs("<pre>\n",out);

 strcpy(pad,"    ");

 while(fgets(line,256,in))
   {
    lineno++;
    if(lineno==10)
       pad[3]=0;
    else if(lineno==100)
       pad[2]=0;
    else if(lineno==1000)
       pad[1]=0;
    else if(lineno==10000)
       pad[0]=0;
    fprintf(out,"<a name=\"line%d\">%d%s|</a> %s",lineno,lineno,pad,html(line,1));
   }

 fputs("</pre>\n",out);
 WriteHTMLPostamble(out,0);

 fclose(in);
 fclose(out);
}


/*++++++++++++++++++++++++++++++++++++++
  Make the input string safe to output as HTML ( not <, >, & or " ).

  char* html Returns a safe HTML string.

  char* c A non-safe HTML string.

  int verbatim Set to true if the text is to be output verbatim ignoring the comment +html+ directives.

  The function can only be called four times in each fprintf() since it returns one of only four static strings.
  ++++++++++++++++++++++++++++++++++++++*/

static char* html(char* c,int verbatim)
{
 static char safe[4][256],*malloced[4]={NULL,NULL,NULL,NULL};
 static int which=0;
 int copy=0,skip=0;
 int i=0,j=0,delta=7,len=256-delta;
 char* ret;

 which=(which+1)%4;
 ret=safe[which];

 safe[which][0]=0;

 if(malloced[which])
   {Free(malloced[which]);malloced[which]=NULL;}

 if(c)
   {
    if(!verbatim)
       i=CopyOrSkip(c,"html",&copy,&skip);

    while(1)
      {
       for(;j<len && c[i];i++)
         {
          if(copy)
            {ret[j++]=c[i]; if(c[i]=='\n') copy=0;}
          else if(skip)
            {               if(c[i]=='\n') skip=0;}
          else
             switch(c[i])
               {
               case 12: /* ^L */
                break;
               case '<':
                strcpy(&ret[j],"&lt;");j+=4;
                break;
               case '>':
                strcpy(&ret[j],"&gt;");j+=4;
                break;
               case '&':
                strcpy(&ret[j],"&amp;");j+=5;
                break;
               case '\n':
                if(j && ret[j-1]=='\n')
                  {
                   strcpy(&ret[j],"<br>");j+=4;
                  }
                ret[j++]=c[i];
                break;
               default:
                ret[j++]=c[i];
               }
          if(c[i]=='\n')
             if(!verbatim)
                i+=CopyOrSkip(c+i,"html",&copy,&skip);
         }

       if(c[i])                 /* Not finished */
         {
          if(malloced[which])
             malloced[which]=Realloc(malloced[which],len+delta+256);
          else
            {malloced[which]=Malloc(len+delta+256); strncpy(malloced[which],ret,(unsigned)j);}
          ret=malloced[which];
          len+=256;
         }
       else
         {ret[j]=0; break;}
      }
   }

 return(ret);
}
